#! /usr/bin/env python
#TS
###############################################################################
#
#  Script name                  : TopspinSMFlash.py
#
#  Dependencies                 : ComDefs.py
#                                 ComFuncs.py
#                                 Logger.py
#                    
#  Objective                    : Script used to login to a Topspin Infiniband
#                                 switch using telnet and update the firmware
#                                 via an FTP server.
#                    
#  Written by                   : Robert K. Christner
#                     
#  Change History               : who -   when   - what
#                                 RKC - 11/29/04 - Initial revision.
#  
#  Copyright IBM Corporation, 2004
#
###############################################################################
"""
   Script used to login to a Topspin switch over telnet and update the
   firmware via a FTP server.
"""
try:
    import time, sys, os, re, getopt, telnetlib, socket
    sys.path.append("..")
    from Common import ComDefs, ComFuncs
    from Common.Logger import Logger
except KeyboardInterrupt:
    sys.exit(1)


# Instantiate Globals
params = ""
logger = ""
lastError = ComDefs.SUCCESS
ipAddr = ComDefs.FALSE
cleanExit = ComDefs.TRUE
standalone = ComDefs.FALSE

def usage():
    '''
        Display the Usage Information to the screen
    '''

    print "Usage: TopspinSMFlash.py [options]"
    print "\nOptions:"
    print "  -i, --ipaddress=IPADDR      The hostname or dotted IP address of the I/O"
    print "                              Module"
    print "  -u, --username=USERNAME     An administrator user name used to log into the"
    print "                              I/O Module"
    print "  -p, --password=PASSWORD     The password of the administrator user name used"
    print "                              to log into the I/O Module"
    print "  -s, --ftpserver=IPADDR      The dotted IP address of the FTP server hosting"
    print "                              the new firmware image"
    print "  -m, --ftpusername=USERNAME  An user name used to log into the FTP server"
    print "  -w, --ftppassword=PASSWORD  The password of the user name used to"
    print "                              log into the FTP server"
    print "  -n, --imagename=FILE        The path and name of the I/O Module firmware image"
    print "                              file on the FTP server"
    print "  -h, --help, -?              Display this (help) message"
    print "  -v, --verbose               Increase the debug output level by one"
    print "  -V, --version               Display the TopspinSMFlash version number and exit"


def parse_args(opts):
    """
       Parse the arguments and options and handle them accordingly
    """

    global params, ipAddr, cleanExit

    for o, a in opts:
        if o in ("-?", "-h", "--help"):
            usage()
            return cleanExit
        elif o in ("-v", "--verbose"):
            params.debugLevel += 1
        elif o in ("-V", "--version"):
            print ComDefs.VERSION
            return cleanExit
        elif o in ("-i", "--ipaddress"):
            ipAddr = a
        elif o in ("-u", "--username"):
            params.io1UserName = a
        elif o in ("-p", "--password"):
            params.io1Password = a
        elif o in ("-s", "--ftpserver"):
            params.ftpIPAddress = a
        elif o in ("-m", "--ftpusername"):
            params.ftpUserName = a
        elif o in ("-w", "--ftppassword"):
            params.ftpPassword = a
        elif o in ("-n", "--imagename"):
            params.io1Filename1 = a
        else:
            usage()
            return cleanExit

    return ComDefs.SUCCESS

def verify_args():
    """
       Verify that the required arguments have been supplied.
    """

    global params, ipAddr, cleanExit

    if ipAddr == "":
        print "The hostname or dotted IP address of the I/O Module was not specified."
        usage()
        return cleanExit
    if params.ftpIPAddress == "":
        print "The dotted IP address of the FTP server was not specified."
        usage()
        return cleanExit
    if params.ftpUserName == "":
        print "The username for the FTP server was not specified."
        usage()
        return cleanExit
    if params.ftpPassword == "":
        print "The password for the FTP server was not specified."
        usage()
        return cleanExit
    if params.io1Filename1 == "":
        print "The path and name of the I/O Module firmware image file on the FTP"
        print "server was not specified."
        usage()
        return cleanExit

    return ComDefs.SUCCESS


def connect(telnet, ioIPAddr, userName, password):
    """
        Connect and authenticate to the switch.
    """

    global params, logger, standalone

    try:
        telnet.open(ioIPAddr)
        logger.log("Connecting to %s" % ioIPAddr)
    except socket.error:
        logger.log("Socket error! Connection refused or timed out to (%s)." % ioIPAddr)
        lastError = ComDefs.TN_SOCKET_ERR
        return lastError
    except KeyboardInterrupt:
        sys.exit(0)

    # look for "login" and enter credentials
    reList = []
    reList.append(re.compile('login'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message recieved:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    logger.log("Connected to Infiniband Switch Module (%s)." % ioIPAddr)
    telnet.write('%s\r' % userName)
    #logger.log("Sending username.")
    logger.log("Sending username (%s)." % userName)
    # look for "Password" and enter credentials
    reList = []
    reList.append(re.compile('Password'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message recieved:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    telnet.write('%s\r' % password)
    logger.log("Sending password.")
    #logger.log("Sending password (%s)." % password)
    # credentials are sent, continue to the next screen
    reList = []
    reList.append(re.compile('>'))
    reList.append(re.compile('login'))
    reList.append(re.compile('Change\ your\ password\ now'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == 1:
        logger.log("Error! Username or Password not accepted.")
        if params.debugLevel > 1:
            print txt
        telnet.close()
        lastError = ComDefs.TN_INVALID_LOGIN
        return lastError
    if index == 2:
        telnet.write('no\r')
    logger.log("Username and Password accepted.")

    lastError = ComDefs.SUCCESS
    return lastError


def update_firmware(telnet, ioIPAddr, userName, password, filename):
    """
        Flash the firmware
    """

    global params, logger, standalone

    # open a telnet session with the switch
    lastError = connect(telnet, ioIPAddr, userName, password)
    if lastError != ComDefs.SUCCESS:
        return lastError

    # credentials are accepted, continue to firmware upgrade
    telnet.write('en\r')
    reList = []
    reList.append(re.compile('#'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message recieved:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    telnet.write('dir image\r')
    reList = []
    reList.append(re.compile('([\S]+\.img)'))
    reList.append(re.compile('#'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message recieved:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    elif index == 0:
        img = mo.group(0)
        # gotta get rid of the extra prompt
        reList = []
        reList.append(re.compile('#'))
        (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
        # now delete the file
        logger.log("Deleting old image file (%s) to make room for a new one." % img, ComDefs.TRUE)
        telnet.write('delete image:%s\r' % img)
        reList = []
        reList.append(re.compile('\*'))
        reList.append(re.compile('Delete\ file'))
        reList.append(re.compile('#'))
        while ComDefs.TRUE:
            (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
            if index == -1:
                logger.log("Error! Telnet timeout.")
                logger.log("Last message recieved:\n%s" % txt)
                telnet.close()
                lastError = ComDefs.GENERAL_UPDATE_ERROR
                return lastError
            elif index == 0:
                sys.stdout.write(".")
                sys.stdout.flush()
            elif index == 1:
                telnet.write("yes\r")
            else:
                print
                logger.log("Old file deleted.")
                break
    
    # downloading firmware
    logger.log("Downloading Infiniband Switch Module firmware image (%s)," % filename, ComDefs.TRUE)
    logger.log("from FTP server (%s)." % params.ftpIPAddress, ComDefs.TRUE)
    #regex = re.compile('[^/\\s]+\.img')
    #file = regex.search(filename).group(0)
    file = os.path.basename(filename)
    logger.log('copy ftp://%s:****@%s/%s image:%s' % (params.ftpUserName, params.ftpIPAddress, filename, file))
    telnet.write('copy ftp://%s:%s@%s/%s image:%s\r' % (params.ftpUserName, params.ftpPassword, params.ftpIPAddress, filename, file))
    reList = []
    reList.append(re.compile('Error: (.*)\.'))
    reList.append(re.compile('%'))
    reList.append(re.compile('successfully'))
    while ComDefs.TRUE:
        (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
        if index == -1:
            logger.log("Error! Telnet timeout.")
            logger.log("Last message recieved:\n%s" % txt)
            telnet.close()
            lastError = ComDefs.GENERAL_UPDATE_ERROR
            return lastError
        elif index == 0:
            logger.log("\bError! %s." % mo.group(1))
            telnet.close()
            # FIXME - use GENERAL_UPDATE_ERROR?
            lastError = ComDefs.IO_FIRMWARE_IMAGE_NOT_FOUND
            return lastError
        elif index == 1:
            sys.stdout.write(".")
            sys.stdout.flush()
        else:
            print
            logger.log("Downloading complete succesfully.")
            break

    # get current image name
    logger.log("Determining the current image name.")
    telnet.write('show boot-config\r')
    reList = []
    reList.append(re.compile('primary-image-source\s:\s(.*)'))
    reList.append(re.compile('Error: (.*)\.'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message recieved:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    elif index == 1:
        logger.log("\bError! %s." % mo.group(1))
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    currimage = mo.group(1).strip()
    logger.log("The current image is %s." % currimage)

    # gotta get rid of the extra prompt
    reList = []
    reList.append(re.compile('#'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    # now install new image
    logger.log("Installing new image (%s)." % file, ComDefs.TRUE)
    telnet.write('install image:%s\r' % file)
    reList = []
    reList.append(re.compile('\*'))
    reList.append(re.compile('Proceed'))
    reList.append(re.compile('Error: (.*)\.'))
    reList.append(re.compile('#'))
    while ComDefs.TRUE:
        (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
        if index == -1:
            logger.log("Error! Telnet timeout.")
            logger.log("Last message recieved:\n%s" % txt)
            telnet.close()
            lastError = ComDefs.GENERAL_UPDATE_ERROR
            return lastError
        elif index == 0:
            sys.stdout.write(".")
            sys.stdout.flush()
        elif index == 1:
            telnet.write("yes\r")
        elif index == 2:
            logger.log("\bError! %s." % mo.group(1))
            telnet.close()
            lastError = ComDefs.GENERAL_UPDATE_ERROR
            return lastError
        else:
            print
            logger.log("New image installed.")
            break

    # delete the image file we just installed 
    # gotta get rid of the extra prompt
    reList = []
    reList.append(re.compile('#'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    # now delete the file
    logger.log("Deleting the image file (%s) we just installed." % file, ComDefs.TRUE)
    telnet.write('delete image:%s\r' % file)
    reList = []
    reList.append(re.compile('\*'))
    reList.append(re.compile('Delete\ file'))
    reList.append(re.compile('#'))
    while ComDefs.TRUE:
        (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
        if index == -1:
            logger.log("Error! Telnet timeout.")
            logger.log("Last message recieved:\n%s" % txt)
            telnet.close()
            lastError = ComDefs.GENERAL_UPDATE_ERROR
            return lastError
        elif index == 0:
            sys.stdout.write(".")
            sys.stdout.flush()
        elif index == 1:
            telnet.write("yes\r")
        else:
            print
            logger.log("Image file deleted.")
            break

    # get the new image name
    logger.log("Determining the new image name.")
    telnet.write('dir image\r')
    reList = []
    reList.append(re.compile('([\S]*/[\S]*)'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message recieved:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    newimage = mo.group(0).strip()
    if newimage == currimage:
        reList = []
        reList.append(re.compile('([\S]*/[\S]*)'))
        (index,mo,txt) = telnet.expect(reList, 10)
        if index == 0:
            newimage = mo.group(0).strip()
    logger.log("The new image name is %s." % newimage)

    # gotta get rid of the extra prompt
    reList = []
    reList.append(re.compile('#'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    # make the new image the boot image
    logger.log("Making the new image the boot image.")
    telnet.write('config\r')
    reList = []
    reList.append(re.compile('#'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message recieved:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    telnet.write('boot-config primary-image-source %s\r' % newimage)
    reList = []
    reList.append(re.compile('Error: (.*)\.'))
    reList.append(re.compile('#'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message recieved:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    elif index == 0:
        logger.log("\bError! %s." % mo.group(1))
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    telnet.write('exit\r')
    reList = []
    reList.append(re.compile('#'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message recieved:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError

    # reboot
    logger.log("Reloading Infiniband Switch Module.", ComDefs.TRUE)
    telnet.write('reload no-failover\r')
    reList = []
    reList.append(re.compile('Save'))
    reList.append(re.compile('Proceed'))
    reList.append(re.compile('Error: (.*)\.'))
    while ComDefs.TRUE:
        (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
        if index == 0:
            telnet.write('yes\r') 
            continue
        elif index == 1:
            telnet.write('yes\r') 
            break
        elif index == 2:
            logger.log("\bError! %s." % mo.group(1))
            telnet.close()
            lastError = ComDefs.GENERAL_UPDATE_ERROR
            return lastError
        elif index == -1:
            logger.log("Error! Telnet timeout.")
            logger.log("Last message recieved:\n%s" % txt)
            telnet.close()
            lastError = ComDefs.GENERAL_UPDATE_ERROR
            return lastError
    
    if standalone:
        logger.log("Infiniband Switch Module firmware succesfully updated.", ComDefs.TRUE)

    telnet.close()
    lastError = ComDefs.SUCCESS
    return lastError


def flash_switch(ioIPAddr, ioBay=1):
    """
       Flash the Topspin switch via interaction with the telnet menu interface.
    """

    global params, logger, standalone, lastError

    if standalone:
        logger.log("Updating Infiniband Switch Module firmware.", ComDefs.TRUE)

    # Get Variables from params class
    if ioBay == 1:
        userName = params.io1UserName
        password = params.io1Password
        filename = params.io1Filename1
    elif ioBay == 2:
        userName = params.io2UserName
        password = params.io2Password
        filename = params.io2Filename1
    elif ioBay == 3:
        userName = params.io3UserName
        password = params.io3Password
        filename = params.io3Filename1
    elif ioBay == 4:
        userName = params.io4UserName
        password = params.io4Password
        filename = params.io4Filename1

    # check a couple params to make sure
    if params.ftpIPAddress == "":
        logger.log("Error! The IP Address of the ftp server was not specified.")
        lastError = ComDefs.FTP_BAD_IP
        return lastError
    if params.ftpUserName == "":
        logger.log("Error! The Username for the ftp server was not specified.")
        lastError = ComDefs.FTP_INVALID_LOGIN
        return lastError
    if userName == "":
        logger.log("Error! Username was not supplied.")
        lastError = ComDefs.TN_INVALID_LOGIN
        return lastError
    if password == "":
        logger.log("Error! Password was not supplied.")
        lastError = ComDefs.TN_INVALID_LOGIN
        return lastError
    if filename == "":
        logger.log("Error! Firmware image filename was not supplied.")
        lastError = ComDefs.IO_INVALID_IMAGE_FILE
        return lastError

    telnet = telnetlib.Telnet()

    # flash the firmware
    lastError = update_firmware(telnet, ioIPAddr, userName, password, filename)
    if lastError != ComDefs.SUCCESS:
        return lastError

    if standalone:
        logger.log("Success! Infiniband Switch Module firmware update complete.", ComDefs.TRUE)

    return ComDefs.SUCCESS


def main():
    """ Main function used to run this script in standalone mode"""

    # Global Variables
    global params, logger, ipAddr, standalone, lastError

    params = ComDefs.parameters()
    standalone = ComDefs.TRUE

    # Parse command line options with getopt()
    try:
        opts, args = getopt.getopt(sys.argv[1:], "?hvVi:u:p:s:m:w:n:",
             ["?", "help", "verbose", "version", 
              "ipaddress=", "username=", "password=",
              "ftpserver=", "ftpusername=", "ftppassword=", "imagename="])

    except getopt.GetoptError:
        # Print usage information and exit
        usage()
        sys.exit(0)
    
    lastError = parse_args(opts)
    if lastError != ComDefs.SUCCESS:
        return lastError

    lastError = verify_args()
    if lastError != ComDefs.SUCCESS:
        return lastError
        
    logger = Logger(params)
    rc = flash_switch(ipAddr)
    if rc != ComDefs.SUCCESS:
        logger.log(ComDefs.GetLastError(rc), ComDefs.TRUE)
    
# This get's invoked if they run this script directly
if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        ComFuncs.keyboardISR(logger)
