#! /usr/bin/env python
###############################################################################
#
#  Script name                  : MMFlash.py
#
#  Dependencies                 : ComDefs.py
#                                 MMComunications.py
#                    
#  Objective                    : Script used to login to the management port (6090)
#                                 of the MM and run dot commands to flash the firmware
#                                 of the MM
#                    
#  Written by                   : Mike Nolterieke
#                     
#  Change History               : who -   when   - what
#                                 MHN - 07/05/04 - Initial revision.
#                                 RKC - 07/30/04 - use getopt instead of optparse
#                                 RKC - 07/30/04 - added logging
#                     
###############################################################################
"""
   Script used to login to the management port (6090) of the MM and run dot commands
   to flash the firmware of the MM
"""
try:
    import sys, time, getopt, urllib, os
    sys.path.append("..")
    from Common import ComDefs, ComFuncs
    from Common.MMCommunications import MMCom
    from Common.Logger import Logger
    from MM import MMInfo
except KeyboardInterrupt:
    sys.exit(1)

# The last error that occurred
lastError = ComDefs.SUCCESS

# The MMCom class
mmCom = ""

params = ""
logger = ""
isTest = ComDefs.FALSE
cleanExit = ComDefs.TRUE

class packetFile:
    build = ""
    path = ""

def updateFirmware(pktFile):
    """ Update the firmware of the MM with the specified packet file
    Return Codes:
        ComDefs.SUCCESS
        ComDefs.MM_PACKET_FILE_NOT_FOUND
        ComDefs.MM_INVALID_PACKET_FILE
        ComDefs.MM_INVALID_PACKET_FILE_HEADER
        Any return code possible from the MMCom.sendPacket function
    """
    global params, mmCom, lastError
    
    # The packet header format
    pktHeader = '\x16\x02\x00\x03'

    # The default return code
    rc = ComDefs.SUCCESS

    # Try to open the packet file
    try:
        packet = open(pktFile, 'rb')
    # If there was an error opening the file
    except:
        return ComDefs.MM_PACKET_FILE_NOT_FOUND

    j = 0
    while ComDefs.TRUE:
        # Making it display some status while sending the packets
        j += 1
        # We were printing too many .'s
        # Only print every 10th dot
        # Only do this if we're verbose
#        if isTest == ComDefs.TRUE:
#            if ((j % 10) == 0):
#                sys.stdout.write(".")
#                sys.stdout.flush()
                
        # Read SLIM packet header
        header = packet.read(6)

        # If the length of the header is 0, we are done
        if (len(header) == 0):
            break

        # If the length of the header is less than 6, the header
        # is invalid
        if (len(header) < 6):
            rc = ComDefs.MM_INVALID_PACKET_FILE_HEADER
            break

        # If the packet file is invalid
        if (header[:4] != pktHeader):
            rc = ComDefs.MM_INVALID_PACKET_FILE
            break

        # Get the slim packet length
        slimPktLen = (ord(header[4]) * 256) + ord(header[5])

        # Pad all odd length packets with a null byte
        #if ((slimPktLen % 2) == 1):
        #    slimPktLen += 1        

        pkt = packet.read(slimPktLen)

        if ((slimPktLen % 2) == 1):
            packet.read(1)

        tail = packet.read(3)
        # If the end of the packet is invalid
        if (tail[2] != '\xC0'):
            rc = ComDefs.MM_INVALID_PACKET_FILE
            break

        # The packet header actually contains the proper dot command
        # to send (8.3.1, 8.3.2, 8.3.3, 8.3.4, etc...) so we do not need to
        # specify the dot command
        # Simply send the entire packet
        rc = mmCom.sendPacket(pkt)
        # If there was an error sending the packet file
        if rc != ComDefs.SUCCESS:
            break

    # Close the file
    packet.close()
    return rc


def SimpleMMFlash(prettyName, packetName):
    """
        Flash the given firmware packet to the MM and provides status
    """

    global params, logger, mmCom, MM_FIRMWARE_PACKETS

    MMInfo.logger = logger
    MMInfo.mmCom = mmCom

    print
    # First copy the firmware to the MM, then kick off the flash.
    logger.log("Copying the %s firmware to the Management Module, please wait." % prettyName, ComDefs.TRUE)
    rc = updateFirmware(packetName)
    if rc != ComDefs.SUCCESS:
        return rc
    else:
        logger.log("Updating the Management Module %s." % prettyName, ComDefs.TRUE)
        flashNotCompleted = ComDefs.TRUE
        isFirstRun = ComDefs.TRUE
        # Loop through a status check to see how the flash is coming along.
        while flashNotCompleted:
            flashStatus = getUpdateStatus()
            if lastError != ComDefs.SUCCESS:
                # If there was an error getting status, there was likely an error with the flash.
                print
                logger.log("There was an error flashing the Management Module %s." % prettyName)
                rc = lastError
                break
            else:
                # Print some status out.
                if isFirstRun:
                    sys.stdout.write("Flash Percent Complete: ")
                    sys.stdout.write(str(ord(flashStatus[1])))
                    sys.stdout.write("%")
                    sys.stdout.flush()
                    isFirstRun = ComDefs.FALSE
                else:
                    # Replace the current status with the new status
                    if flashStatus[0] == chr(255):
                        sys.stdout.write('\b\b\b')
                        sys.stdout.write(str(ord(flashStatus[1])))
                        sys.stdout.write("%")
                        sys.stdout.flush()
                    else:
                        # We have finished, let the status reflect that fact
                        sys.stdout.write('\b\b\b')
                        sys.stdout.write("100%")
                        sys.stdout.flush()
                        flashNotCompleted = ComDefs.FALSE
                        print
                        logger.log("Successfully updated the Management Module %s." % prettyName, ComDefs.TRUE)
                        break
            # Wait a couple seconds and get status again
            time.sleep(2)

    return rc


def FullMMFlash(mmpackets):
    """
        Flashes all the firmware packets to the MM and handles retries
    """

    global params, logger, mmCom

    # Get the most recent information from the MM
    mmfirmware = MMInfo.MMStatus()
    mmfirmware.gather()
    mmtable = mmfirmware.getStatusTable()
    activeBay = mmfirmware.returnActiveBay()
    inactiveBay = mmfirmware.returnInactiveBay()

    # Make sure we actually need to update anything
    # We don't have the logic to backlevel
    if mmtable.mm[activeBay].mainbuild > mmpackets[ComDefs.MM_MAIN].build or \
            mmtable.mm[activeBay].remotebuild > mmpackets[ComDefs.MM_REMOTE].build:
        logger.log(ComDefs.GetLastError(ComDefs.MM_NO_BACKLEVEL), ComDefs.TRUE)
        return ComDefs.MM_NO_BACKLEVEL

    # We don't need to update since the levels are already the same
    if mmtable.mm[activeBay].mainbuild == mmpackets[ComDefs.MM_MAIN].build and \
            mmtable.mm[activeBay].remotebuild == mmpackets[ComDefs.MM_REMOTE].build:
        if mmtable.mm[inactiveBay].installed == ComDefs.FALSE:
            logger.log(ComDefs.MM_EQUAL, ComDefs.TRUE)
            return ComDefs.SUCCESS
        elif mmtable.mm[inactiveBay].mainbuild == mmpackets[ComDefs.MM_MAIN].build and \
                mmtable.mm[inactiveBay].remotebuild == mmpackets[ComDefs.MM_REMOTE].build:
            logger.log(ComDefs.MM_EQUAL, ComDefs.TRUE)
            return ComDefs.SUCCESS


    # Count the number of failures, so we know when we've tried too many times
    failures = 0

    while failures < ComDefs.MAX_MM_FAILURES:
        mmfirmware.gather()
        mmtable = mmfirmware.getStatusTable()
        activeBay = mmfirmware.returnActiveBay()
        inactiveBay = mmfirmware.returnInactiveBay()

        logger.log("The Management Module in bay %d is primary." % activeBay)

        if params.debugLevel > 1:
            mmfirmware.toString()

        # Run the guantlet backwards to see what the next step is
        # Have we committed the update by restarting the MM a second time?
        if mmtable.mm[activeBay].committed == ComDefs.FALSE:
            # Have we restarted the MM to apply the update?
            if mmtable.mm[activeBay].restarted == ComDefs.FALSE:
                # Is the Main App up to date?
                if mmtable.mm[activeBay].mainstatus == ComDefs.FALSE:
                    # Is the Remote App up to date?
                    if mmtable.mm[activeBay].remotestatus == ComDefs.FALSE:
                        # Update the Remote Control firmware
                        logger.log("Updating the Management Module %s firmware from %s to %s" % 
                                  (ComDefs.MM_REMOTE, mmtable.mm[activeBay].remotebuild, 
                                   mmpackets[ComDefs.MM_REMOTE].build), ComDefs.TRUE)
                        rc = SimpleMMFlash(ComDefs.MM_REMOTE, mmpackets[ComDefs.MM_REMOTE].path)
                        # Handle a failure
                        if rc != ComDefs.SUCCESS:
                            logger.log(ComDefs.GetLastError(rc))
                            failures += 1
                            # If the failure was socket related we need to reconnect before continuing
                            if rc == ComDefs.MM_SOCKET_TIMEOUT or rc == ComDefs.MM_SOCKET_ERR or \
                                    rc == ComDefs.MM_UNKNOWN_ERR:
                                logger.log("Waiting for the Management Module to restart before reconnecting to it.", ComDefs.TRUE)
                                time.sleep(60)
                                rc = connectMM(ComDefs.TRUE)
                                if rc != ComDefs.SUCCESS:
                                    # If this connect fails, then we don't want to retry the flash again.
                                    return rc
                            continue
                        # The Remote App has now been updated, reboot if necessary
                        else:
                            mmfirmware.setRemoteStatus(activeBay)
                            # If the current MM build is > 79A and < 82A, we have to reboot between the flashes
                            if mmtable.mm[activeBay].mainbuild >= ComDefs.FORCE_REBOOT_MM_MAIN_CODELEVEL_LOWER and \
                                    mmtable.mm[activeBay].mainbuild <= ComDefs.FORCE_REBOOT_MM_MAIN_CODELEVEL_UPPER:
                                print
                                logger.log("Restarting the Management Module to commit the flash operation.", ComDefs.TRUE)
                                rebootMM()
                                rc = connectMM(ComDefs.TRUE)
                                if rc != ComDefs.SUCCESS:
                                    # If this connect fails, then we don't want to retry the flash again.
                                    return rc
                    # The Remote App is up to date
                    else:
                        # Update the Main App firmware
                        logger.log("Updating the Management Module %s from %s to %s" % 
                                  (ComDefs.MM_MAIN, mmtable.mm[activeBay].mainbuild, 
                                   mmpackets[ComDefs.MM_MAIN].build), ComDefs.TRUE)
                        rc = SimpleMMFlash(ComDefs.MM_MAIN, mmpackets[ComDefs.MM_MAIN].path)
                        # Handle a failure
                        if rc != ComDefs.SUCCESS:
                            logger.log(ComDefs.GetLastError(rc))
                            failures += 1
                            # If the failure was socket related we need to reconnect before continuing
                            if rc == ComDefs.MM_SOCKET_TIMEOUT or rc == ComDefs.MM_SOCKET_ERR or \
                                    rc == ComDefs.MM_UNKNOWN_ERR:
                                logger.log("Waiting for the Management Module to restart before reconnecting to it.", ComDefs.TRUE)
                                time.sleep(60)
                                rc = connectMM(ComDefs.TRUE)
                                if rc != ComDefs.SUCCESS:
                                    # If this connect fails, then we don't want to retry the flash again.
                                    return rc
                            continue
                        # The Main App has now been updated, reboot
                        else:
                            mmfirmware.setMainStatus(activeBay)
                            mmfirmware.setRestarted(activeBay)
                            # Reboot and update the status
                            logger.log("Restarting the Management Module to commit the flash operation.", ComDefs.TRUE)
                            rebootMM()
                            rc = connectMM(ComDefs.TRUE)
                            if rc != ComDefs.SUCCESS:
                                # If this connect fails, then we don't want to retry the flash again.
                                return rc
                # The Main App has been updated, but the MM has yet to be rebooted
                else:
                    mmfirmware.setRestarted(activeBay)
                    # Reboot and update the status
                    logger.log("Restarting the Management Module to commit the flash operation.", ComDefs.TRUE)
                    rebootMM()
                    rc = connectMM(ComDefs.TRUE)
                    if rc != ComDefs.SUCCESS:
                        # If this connect fails, then we don't want to retry the flash again.
                        return rc
            # The update has been restarted for the active bay, but not committed
            else:
                # Reboot one extra time just to be sure everything is settled
                mmfirmware.setCommitted(activeBay)
                print
                logger.log("Rebooting the Management Module again to ensure all updates are settled.", ComDefs.TRUE)
                logger.log("Please be patient, this can take up to 2 minutes.", ComDefs.TRUE)
                rebootMM()
                rc = connectMM()
                if rc != ComDefs.SUCCESS:
                    return rc
        # The update has been committed for the active bay
        else:
            # Check if there is a redundant MM installed
            if mmtable.mm[inactiveBay].installed == ComDefs.TRUE:
                # Check if the redundant needs flashing (we allow a backlevel here if necessary)
                if mmtable.mm[inactiveBay].committed == ComDefs.TRUE:
                    logger.log("The redundant Management Module has been committed.")
                    # ALL DONE
                    break
                elif mmtable.mm[inactiveBay].mainstatus == ComDefs.TRUE and \
                        mmtable.mm[inactiveBay].remotestatus == ComDefs.TRUE:
                    logger.log("The redundant Management Module is already up to date.")
                    # ALL DONE
                    break
                # The redundant is not up to date yet
                else:
                    # FIXME - this doesn't work all the time, so we are just going to manually
                    # flash the redundant and not rely on the MM to do it automatically
                    #
                    # Failover to the redundant MM unless we are waiting for an autoflash
                    #if mmtable.mm[activeBay].mainbuild >= ComDefs.SUPPORTED_MM_MAIN_CODELEVEL and \
                    #        mmtable.mm[inactiveBay].mainbuild >= ComDefs.SUPPORTED_MM_MAIN_CODELEVEL:
                        # The primary MM will go ahead and flash the redundant MM in time
                    #    mmtable.mm[inactiveBay].restarted = ComDefs.TRUE
                    #    mmtable.mm[inactiveBay].committed = ComDefs.TRUE
                    #    break
                    # We have to failover and flash the redundant
                    #else:
                    logger.log("Switching over to the Management Module in order to update it.", ComDefs.TRUE)
                    logger.log("Please be patient, this can take up to 2 minutes.", ComDefs.TRUE)
                    rebootMM(ComDefs.TRUE)
                    rc = connectMM()
                    if rc != ComDefs.SUCCESS:
                        return rc
                    # We are now the active MM so restart the loop
                    continue
            # There is no redundant MM in the system
            else:
                logger.log("There is not a redundant Management Module installed in this system.")
                # ALL DONE
                break

    if failures >= ComDefs.MAX_MM_FAILURES:
        return ComDefs.MM_UPDATE_RETRY_LIMIT_EXCEEDED

    return rc


def getUpdateStatus():
    """ Get the current status of the firmware update
    Returns:
        The response from the MM that is the status of the firmware update
        or an empty string if there was an error
    
        The first byte of the response is the status and the second byte of the
        response is the percentage complete

    If an empty string is returned, the error code will be stored in the lastError
    global variable
    """
    global params, mmCom, logger, lastError
    
    flashStatus = ""

    # Send the 8.3.5 command to the MM to check the update status
    lastError = mmCom.sendCmd('R', '\x08\x03\x05', '')
    # If there was an error 
    if lastError != 0:
        return flashStatus
    # Get the response from the command
    flashStatus = mmCom.getResponse()
    # Reset the lastError variable and return the status
    lastError = ComDefs.SUCCESS
    return flashStatus

def getPacketFilePaths(mmpackets):
    """
        Determines the packet file names and associates them with the type of firmware
    """
    global params, mmCom, logger

    # filename 1
    # BladeScanner will set this empty because a Boot ROM is no longer required
    if params.mmFilename1 != "":
        type = getPacketType(params.mmFilename1)
        if type == "Unknown":
            return ComDefs.MM_PACKET_FILE_NOT_FOUND
        # We do not want to flash the Boot ROM even if it exists
        if type != ComDefs.MM_BOOT:
            mmpackets[type].path = params.mmFilename1

    # filename 2
    type = getPacketType(params.mmFilename2)
    if type == "Unknown":
        return ComDefs.MM_PACKET_FILE_NOT_FOUND
    # We do not want to flash the Boot ROM even if it exists
    if type != ComDefs.MM_BOOT:
        mmpackets[type].path = params.mmFilename2

    # filename 3
    type = getPacketType(params.mmFilename3)
    if type == "Unknown":
        return ComDefs.MM_PACKET_FILE_NOT_FOUND
    # We do not want to flash the Boot ROM even if it exists
    if type != ComDefs.MM_BOOT:
        mmpackets[type].path = params.mmFilename3

    if mmpackets[ComDefs.MM_MAIN].path == "":
        logger.log("%s firmware file could not be found." % ComDefs.MM_MAIN, ComDefs.TRUE)
        return ComDefs.MM_PACKET_FILE_NOT_FOUND
    elif mmpackets[ComDefs.MM_REMOTE].path == "":
        logger.log("%s firmware file could not be found." % ComDefs.MM_REMOTE, ComDefs.TRUE)
        return ComDefs.MM_PACKET_FILE_NOT_FOUND

    return ComDefs.SUCCESS


def getPacketType(pktFile):
    """
        Determines the firmware type given the packet file
    """
    global params, mmCom, logger

    try: 
        packet = open(pktFile, 'rb')
    except IOError:
        logger.log("Cannot open Management Module packet file %s for reading." % pktFile)
        return "Unknown"

    packet.seek(139);
    type = packet.read(2)
    packet.close()

    if type == "BR":
        filetype = ComDefs.MM_BOOT
    elif type == "MN":
        filetype = ComDefs.MM_MAIN
    elif type == "RG":
        filetype = ComDefs.MM_REMOTE
    else:
        logger.log("Could not determine packet type for Management Module packet file %s." % pktFile)
        return "Unknown"

    return filetype


def setPacketRevs(mmpackets):
    """
        Sets the values in the global dictionary to contain the packet build IDs
    """
    global params, mmCom, logger

    rev = getPacketRev(mmpackets[ComDefs.MM_MAIN].path)
    if rev == ComDefs.GENERAL_UPDATE_ERROR:
        return rev
    mmpackets[ComDefs.MM_MAIN].build = rev

    rev = getPacketRev(mmpackets[ComDefs.MM_REMOTE].path)
    if rev == ComDefs.GENERAL_UPDATE_ERROR:
        return rev
    mmpackets[ComDefs.MM_REMOTE].build = rev

    return ComDefs.SUCCESS


def getPacketRev(pktFile):
    """
        Determines the firmware revision given the packet file
    """
    global params, mmCom, logger

    try:
        packet = open(pktFile, 'rb')
    except IOError:
        logger.log("Cannot open packet file %s for reading." % pktFile)
        return ComDefs.GENERAL_UPDATE_ERROR

    packet.seek(109);
    version = packet.read(7)
    packet.close()

    return version


def comparePacketRevs(mmpackets):
    """
        Compare the three MM packet files to make sure they are the same version 
        and return SUCCESS or ERROR

        getPacketFilePaths() must be called before this function
    """
    global params, mmCom, logger

    # Get Main App Build ID
    rev1 = getPacketRev(mmpackets[ComDefs.MM_MAIN].path)
    if rev1 == ComDefs.GENERAL_UPDATE_ERROR:
        return ComDefs.GENERAL_UPDATE_ERROR

    # Get Remote Control Build ID
    rev2 = getPacketRev(mmpackets[ComDefs.MM_REMOTE].path)
    if rev2 == ComDefs.GENERAL_UPDATE_ERROR:
        return ComDefs.GENERAL_UPDATE_ERROR

    # Compare the results
    if rev1[4:] != rev2[4:]:
        logger.log("The specified Management Module firmware files are not all the same version.")
        return ComDefs.GENERAL_UPDATE_ERROR

    return ComDefs.SUCCESS


def rebootMMweb(doFailOver = ComDefs.FALSE):
    """ Reboot the Management Module via the web interface
    Specify ComDefs.TRUE for doFailOVer if you want the MM to fail over to the redundant MM
    This is only supported on MM firmware levels at 52 or higher
    
    Return Codes:
        ComDefs.SUCCESS
        ComDefs.MM_SOCKET_ERR
    """
    global logger, params, mmCom

    rc = ComDefs.SUCCESS
    if doFailOver:
        # Send the reboot command and let it fail over if there is a redundant MM
        try:
            u = urllib.urlopen("http://%s:%s@%s/private/failover_mm" % (params.mmUserName, params.mmPassword, params.mmIPAddress))
            u.close()
        except:
            logger.log("Failover of Management Module failed.  Could not connect to web interface.")
            rc = ComDefs.MM_SOCKET_ERR
    else:
        # Send the reboot command and do not let it fail over if there is a redundant MM
        try:
            u = urllib.urlopen("http://%s:%s@%s/private/restart_asm" % (params.mmUserName, params.mmPassword, params.mmIPAddress))
            u.close()
        except:
            logger.log("Restart of Management Module failed.  Could not connect to web interface.")
            rc = ComDefs.MM_SOCKET_ERR

    # Sleep to let the MM come back to life
    if rc == ComDefs.SUCCESS:
        time.sleep(40)
        clearArp()

    return rc


def handleReboot():
    """
    Reboot the Management Module of the forceMMreboot parameter is passed,
    and handle calling connectMM again if reboot succeeds.
    """
    global params, logger, mmCom

    if params.forceMMreboot:
        logger.log("Rebooting the Management Module.", ComDefs.TRUE)
        logger.log("Please be patient, this can take up to 2 minutes.", ComDefs.TRUE)
        rc = rebootMMweb()
        if rc != ComDefs.SUCCESS:
            logger.log("Error! Reboot of the Management Module failed.", ComDefs.TRUE)
            return ComDefs.MM_SOCKET_ERR
        else:
            return connectMM()
    else:
        print
        logger.log("Some management applications can hold open port 6090 of the Management Module.", ComDefs.TRUE)
        logger.log("In this situation to force a reboot of the Management Module via", ComDefs.TRUE)
        logger.log("the web interface, which will release port 6090, please set the Force Reboot", ComDefs.TRUE)
        logger.log("option for the Management Module using BladeScanner.", ComDefs.TRUE)
        return ComDefs.MM_SOCKET_ERR


def clearArp():
    """
    clears the entry of the Management Module in the arp cache
    """
    global params, logger

    logger.log("Removing Management Module IP Address from arp table.")

    if os.name == "posix":
        os.system("arp -d %s 2> /dev/null" % params.mmIPAddress)
    else:
        os.system("arp -d 2> NUL")

    return


def connectMM(reboot=ComDefs.FALSE, maxRetries=ComDefs.MAX_MM_RETRIES):
    """
    Try to connect to the Management Module.  If there is a socket error, 
    retry "maxRetries" times.  Then reboot once, and start the retry period
    again.
    """
    global params, logger, mmCom

    logger.log("Connecting to the BladeCenter Management Module.", ComDefs.TRUE)

    rc = mmCom.open(ComDefs.TRUE, params.mmIPAddress, 6090, 5, params.mmUserName, params.mmPassword)
    
    # Because applications (like Director) access the RSA, RSA2, and
    # MM on port 6090 (which the MM uses), we loop through the MM
    # connection if we can't access port 6090.  We quit after
    # attempt number "maxRetries".
    if rc == ComDefs.MM_SOCKET_ERR:
        mmAttempt = 1
        while mmAttempt <= maxRetries:
            logger.log("Socket error connecting to Management Module (%s)" % params.mmIPAddress)
            time.sleep(10)
            logger.log("Retry #%d of %d" % (mmAttempt, maxRetries), ComDefs.TRUE)
            clearArp()
            rc = mmCom.open(ComDefs.TRUE, params.mmIPAddress, 6090, 5, params.mmUserName, params.mmPassword)
            if rc != ComDefs.MM_SOCKET_ERR:
                break
            mmAttempt += 1
            
        if rc == ComDefs.MM_SOCKET_ERR and mmAttempt > maxRetries:
            logger.log("Maximum retries exceeded.  Unable to connect to the Management Module.", ComDefs.TRUE)
            logger.log("Verify network settings and the mmipaddr entry of the response file.", ComDefs.TRUE)
            if reboot:
                rc = handleReboot()

    # wait a few seconds before trying to access the MM
    if rc == ComDefs.SUCCESS:
        time.sleep(5)

    return rc


def rebootMM(doFailOver=ComDefs.FALSE):
    """ Reboot the MM
    Specify ComDefs.TRUE for doFailOVer if you want the MM to fail over to the redundant MM
    
    Return Codes:
        ComDefs.SUCCESS
        Any return code possible from the MMCom.sendCmd function
    """
    global params, mmCom, logger


    rc = ComDefs.SUCCESS
    if doFailOver:
        # Send the reboot command and let it fail over if there is a redundant MM
        # give any previous commands time to complete before rebooting
        time.sleep(5)
        rc = mmCom.sendCmd('W', '\x07\x04', '')
    else:
        # Send the reboot command and do not let it fail over if there is a redundant MM
        # Disable failover for 10 seconds
        rc = mmCom.sendCmd('W', '\x16\x02\x03', '\x0A\x00')
        # give any previous commands time to complete before rebooting
        time.sleep(5)
        rc = mmCom.sendCmd('W', '\x07\x04\x01', '')

    # Sleep to let the MM come back to life
    if rc == ComDefs.SUCCESS:
        logger.log("Waiting for the Management Module to restart before reconnecting to it.", ComDefs.TRUE)
        time.sleep(40)
        clearArp()

    return rc


def usage():
    '''
        Display the Usage Information to the screen
    '''

    print "Usage: MMFlash.py [options]"
    print "\nOptions:"
    print "  -i, --ipaddress=IPADDR   The hostname or dotted IP address of the Management"
    print "                           Module"
    print "  -u, --username=USERNAME  An administrator user name used to log into the"
    print "                           Management Module"
    print "  -p, --password=PASSWORD  The password of the administrator user name used to"
    print "                           log into the Management Module"
    print "  -h, --help, -?           Display this (help) message"
    print "  -v, --verbose            Increase the debug output level by one"
    print "  -V, --version            Display the MMFlash version number and exit"


def parse_args(opts):
    """
       Parse the arguments and options and handle them accordingly
    """

    global isTest, cleanExit, params, logger, mmCom

    for o, a in opts:
        if o in ("-?", "-h", "--help"):
            usage()
            return cleanExit
        elif o in ("-v", "--verbose"):
            params.debugLevel += 1
        elif o in ("-V", "--version"):
            print ComDefs.VERSION
            return cleanExit
        # Hidden test parameter, specify -t to use our known ip address, user name,
        # and passwords to test this script
        elif o in ("-t", "--test"):
            isTest = ComDefs.TRUE
        elif o in ("-i", "--ipaddress"):
            params.mmIPAddress = a
        elif o in ("-u", "--username"):
            params.mmUserName = a
        elif o in ("-p", "--password"):
            params.mmPassword = a
        else:
            usage()
            return cleanExit

    return ComDefs.SUCCESS


def verify_args():
    """
       Verify that the required arguments have been supplied.
    """

    global cleanExit, params, mmCom, logger

    if params.mmIPAddress == "":
        print "The hostname or dotted IP address of the Management Module was not specified."
        usage()
        return cleanExit

    return ComDefs.SUCCESS


def main():
    ''' Main function used to run this script in standalone mode'''
    # Define using a global variable
    global params, mmCom, lastError, logger, packetFile

    # Instanciate the class
    params = ComDefs.parameters()

    # Parse command line options with getopt()
    try:
        opts, args = getopt.getopt(sys.argv[1:], "?htvVi:u:p:",
             ["?", "help", "test", "verbose", "version", "ipaddress=", "username=", "password="])

    except getopt.GetoptError:
        # Print usage information and exit
        usage()
        sys.exit(2)

    lastError = parse_args(opts)
    if lastError != ComDefs.SUCCESS:
        return lastError

    # If we are in test mode (-t parameter), use our defaults and ignore the rest
    # of the parameters
    if isTest:
        # Populate the class
        params.mmIPAddress = ""
        params.mmUserName = ""
        params.mmPassword = ""
    else:
        lastError = verify_args()
        if lastError != ComDefs.SUCCESS:
            return lastError

    # Uncomment this line to force a reboot of the MM via the web interface
    # if a connection cannot be made through the socket interface. This is
    # useful when IBM Director is in your environment, and is holding open
    # the 6090 port of the MM
    #params.forceMMreboot = ComDefs.TRUE

    # Hard code the location of the packet files to be CWD
    params.mmFilename1 = ""
    params.mmFilename2 = "cnetmnus.pkt"
    params.mmFilename3 = "cnetrgus.pkt"
        
    logger = Logger(params)
    mmCom = MMCom(params.debugLevel)
    mmCom.logger = logger

    # The MM firmware packets
    mmpackets = {ComDefs.MM_MAIN:packetFile(),ComDefs.MM_REMOTE:packetFile()}

    # Populate mmpackets with the path of each packet
    rc = getPacketFilePaths(mmpackets)
    if rc != ComDefs.SUCCESS:
        logger.log(ComDefs.GetLastError(rc), ComDefs.TRUE)
        return rc

    # Populate mmpackets with the build of each packet
    rc = setPacketRevs(mmpackets)
    if rc != ComDefs.SUCCESS:
        logger.log(ComDefs.GetLastError(rc), ComDefs.TRUE)
        return rc
    
    logger.log("Updating Management Module", ComDefs.TRUE)
    # Connect to the MM port 6090
    rc = connectMM(ComDefs.TRUE, 3)
    if rc != ComDefs.SUCCESS:
        logger.log("There was an error connecting or logging into the MM!", ComDefs.TRUE)
        logger.log(ComDefs.GetLastError(rc), ComDefs.TRUE)
        sys.exit(1)
    else:
        logger.log("Successfully connected and logged into the MM!")

    # Call this section instead of the next if you want to flash the MM
    # and take into account redundant MMs and any other intelligence needed to 
    # handle different possible levels of MM firmware
    #rc = FullMMFlash(mmpackets)

    # Just flash the files we find locally without any checks/workarounds
    # Comment this section out if you want to use FullMMFLash instead.
    for prettyName, packet in mmpackets.iteritems():
        rc = SimpleMMFlash(prettyName, packet.path)
        if rc != ComDefs.SUCCESS:
            break

    if rc != ComDefs.SUCCESS:
        logger.log("One or more of the flashes did not complete successfully!", ComDefs.TRUE)
        logger.log(ComDefs.GetLastError(rc), ComDefs.TRUE)
    else:
        # If all of the flashes worked, reboot the MM but do not let it fail over
        print
        logger.log("All updates completed successfully!", ComDefs.TRUE)
        logger.log("Rebooting the Management Module.", ComDefs.TRUE)
        time.sleep(5)
        mmCom.sendCmd('W', '\x07\x04\x01', '')

    # Close the connection to the MM
    mmCom.close()
    sys.exit(lastError)
    
# This get's invoked if they run this script directly
if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        ComFuncs.keyboardISR(logger)
